import sys
import os
from argparse import ArgumentParser
import secrets

from blockerbugs import db, app
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.release import Release
from blockerbugs.util.bug_sync import BugSync
from blockerbugs.util.update_sync import UpdateSync
from blockerbugs.util.bz_interface import BlockerBugs
import blockerbugs.util.discussion_sync as discussion_sync
import blockerbugs.util.pagure_bot as pagure_bot
from blockerbugs.models.bug import Bug
from alembic.config import Config as al_Config
from alembic import command as al_command
from alembic import script
from alembic.migration import MigrationContext
from sqlalchemy.engine import reflection


def get_alembic_config():
    # the location of the alembic ini file and alembic scripts changes when
    # installed via package
    if os.path.exists("./alembic.ini"):
        alembic_cfg = al_Config("./alembic.ini")
    else:
        alembic_cfg = al_Config("/usr/share/blockerbugs/alembic.ini",
                                ini_section='alembic-packaged')
    return alembic_cfg

def init_db(args):
    initialize_db(destructive=args.destructive)

def initialize_db(destructive=False):
    alembic_cfg = get_alembic_config()

    if destructive:
        print("Initializing Database")
        db.drop_all()
        db.create_all()
        print("Initializing alembic version")
        al_command.stamp(alembic_cfg, "head")
        return True

    # check whether the table 'milestone' exists
    # if it does not, we assume that the database is empty
    insp = reflection.Inspector.from_engine(db.engine)
    table_names = insp.get_table_names()
    if 'milestone' not in table_names:
        print(" - Creating tables")
        db.create_all()
        print(" - Stamping alembic's current version to 'head'")
        al_command.stamp(alembic_cfg, "head")
        return True

    # check to see if the db has already been initialized by checking for an
    # alembic revision
    directory = script.ScriptDirectory.from_config(alembic_cfg)
    context = MigrationContext.configure(db.engine.connect())
    current_rev = context.get_current_revision()
    if not current_rev:
        print("Couldn't determine alembic version in db...")
        return False
    if set(context.get_current_heads()) == set(directory.get_heads()):
        print("Database already initialized and at the latest rev %s - not re-initializing" % current_rev)
        return True
    else:
        print("Database is behind alembic scripts, upgrading...")
        upgrade_db(None)
        return True

    # If we hit this point, database is probably screwed
    return False


def upgrade_db(args):
    print("Upgrading Database to Latest Revision")
    alembic_cfg = get_alembic_config()
    al_command.upgrade(alembic_cfg, "head")


def add_release(args):
    number = args.release
    existing_release = Release.query.filter_by(number=number).count()
    if existing_release > 0:
        print("Release %s already exists, not adding" % number)
    else:
        print("Adding release %s" % number)
        release_num = int(number)
        new_release = Release(release_num)
        db.session.add(new_release)
        db.session.commit()


def add_milestone(args):
    number = args.release
    version = args.milestone
    blocker = args.blocker
    fe = args.fe

    existing_milestone = Milestone.query.filter_by(version=version,
                                                   blocker_tracker=blocker,
                                                   fe_tracker=fe).count()

    current_milestones = Milestone.query.filter_by(active=True).count()
    if existing_milestone > 0:
        print("Milestone already exists, not adding: %s-%s (%s, %s)" % (number, version, blocker, fe))
    else:
        print("Adding milestone: %s-%s (%s, %s)" % (number, version, blocker, fe))
        release_num = int(number)
        release = Release.query.filter_by(number=release_num).first()
        if not release:
            add_release(release)
            release = Release.query.filter_by(number=release_num).first()

        new_milestone = Milestone(release, version, int(blocker), int(fe),
                                  '%d-%s' % (release_num, version))

        # only active releases are synced but since this version is being added,
        # it's a good assumption to make it active.
        new_milestone.active = True

        # there should be only one current version, the new version as current if
        # there are no current releases defined
        if current_milestones <= 0:
            new_milestone.current = True
        db.session.add(new_milestone)
        db.session.commit()


def generate_config(args):
    dburi = args.dburi

    config_filename = os.path.abspath('conf/settings.py')
    if os.path.exists(config_filename):
        print("configuration file %s already exists, exiting" % config_filename)
        sys.exit(0)

    secret_key = secrets.token_hex()

    config_file = open(config_filename, 'w')
    config_file.write("SECRET_KEY = '%s'\n" % secret_key)
    config_file.write("SQLALCHEMY_DATABASE_URI = '%s'\n" % dburi or '')
    # leave places for the other config values generally needed
    config_file.write("FAS_ADMIN_GROUP = ''\n")
    config_file.write("FAS_USER = ''\n")
    config_file.write("FAS_PASSWORD = ''\n")
    config_file.write("PAGURE_REPO_TOKEN = ''\n")
    config_file.close()


def check_blockers(bzurl):
    bzinterface = BlockerBugs(url=bzurl)
    active_milestones = Milestone.query.filter_by(active=True).all()
    for milestone in active_milestones:
        missing_bugs = []
        buglist = bzinterface.do_query(milestone.blocker_tracker)
        for bug in buglist:
            bugid = bug.bug_id
            new_bug = Bug.query.filter_by(bugid=bugid).first()
            if not new_bug:
                missing_bugs.append(bugid)

        print("========================================")
        print("Fedora %d %s Missing Blocker Bugs" % (milestone.release.number, milestone.version))
        print("========================================")
        if len(missing_bugs) > 0:
            for bug in missing_bugs:
                print("  * %d" % bug)
        else:
            print("  No missing bugs found")


def sync_bugs(args):
    fullsync = args.full
    docheck = args.check

    # grab the URL from app config so that it's syncing from the right place
    bzurl = app.config['BUGZILLA_XMLRPC']

    current_milestone = Milestone.query.filter_by(active=True).first()
    if not current_milestone:
        sys.stderr.write('BugSync ERROR: No active releases found!')
        sys.exit(1)

    sync = BugSync(db, url=bzurl)

    sync.update_active(full_sync=fullsync)

    if docheck:
        print("Checking bug sync status")
        check_blockers(bzurl)


def sync_updates(args):
    active_releases = Release.query.filter_by(active=True).all()
    update_sync = UpdateSync(db)
    for release in active_releases:
        update_sync.sync_updates(release)


def sync_discussions(args):
    discussion_sync.sync_discussions()


def sync(args):
    sync_bugs(args)
    sync_updates(args)
    sync_discussions(args)


def recreate_discussion(args):
    bugid = int(args.bugid)
    bug = Bug.query.filter_by(bugid=bugid).first()

    if not bug:
        app.logger.error('No bug %d in database.' % bugid)
        sys.exit(1)

    discussion_sync.recreate_discussion(bugid)


def update_discussion(args):
    ticket_id = int(args.ticket_id)
    pagure_bot.webhook_handler(ticket_id)


def close_inactive_discussions(args):
    discussion_sync.close_discussions_inactive_releases(args.dryrun)


def main():
    parser = ArgumentParser()

    subparsers = parser.add_subparsers(dest='command', metavar="<COMMAND>")

    init_db_parser = subparsers.add_parser('init_db', help='Initialize DB')
    init_db_parser.add_argument('-rm', '--destructive', action='store_true',
                                dest='destructive', default=False,
                                help='Force database recreation (will ERASE your DB!!!)')
    init_db_parser.set_defaults(func=init_db)

    add_milestone_parser = subparsers.add_parser('add_milestone',
                                                 help='Add milestone')
    add_milestone_parser.add_argument('-r', '--release',
                                      dest='release', help='release number')
    add_milestone_parser.add_argument('-m', '--milestone',
                                      dest='milestone', help='Milestone name')
    add_milestone_parser.add_argument('-b', '--blocker',
                                      dest='blocker', help='blocker tracking bug id')
    add_milestone_parser.add_argument('-a', '--accepted',
                                      dest='fe', help='FE tracking bug id')
    add_milestone_parser.set_defaults(func=add_milestone)

    add_release_parser = subparsers.add_parser('add_release', help='Add release')
    add_release_parser.add_argument('-r', '--release',
                                    dest='release', help='release number')
    add_release_parser.set_defaults(func=add_release)

    generate_config_parser = subparsers.add_parser('generate_config',
                                                   help='Generate config')
    generate_config_parser.add_argument('-d', '--dburi',
                                        dest='dburi', help='dburi to use')
    generate_config_parser.set_defaults(func=generate_config)

    sync_parser = subparsers.add_parser('sync',
                                        help='Synchronize all')
    sync_parser.add_argument('-f', '--full',
                             action='store_true', default=False,
                             help='Force full sync')
    sync_parser.add_argument('-c', '--check',
                             action='store_true', default=False,
                             help='Force check for missing blocker bugs after sync')
    sync_parser.set_defaults(func=sync)

    sync_bugs_parser = subparsers.add_parser('sync-bugs',
                                             help='Synchronize bugs only')
    sync_bugs_parser.add_argument('-f', '--full',
                                  action='store_true', default=False,
                                  help='Force full sync')
    sync_bugs_parser.add_argument('-c', '--check',
                                  action='store_true', default=False,
                                  help='Force check for missing blocker bugs after sync')
    sync_bugs_parser.set_defaults(func=sync_bugs)

    sync_updates_parser = subparsers.add_parser('sync-updates',
                                                help='Synchronize updates only')
    sync_updates_parser.add_argument('-f', '--full',
                                     action='store_true', default=False,
                                     help='Deprecated, no longer does anything')
    sync_updates_parser.set_defaults(func=sync_updates)

    sync_discussions_parser = subparsers.add_parser('sync-discussions',
                                                    help='Synchronize discussions only')
    sync_discussions_parser.set_defaults(func=sync_discussions)

    upgrade_db_parser = subparsers.add_parser('upgrade_db',
                                              help='Upgrade DB to newer revision')
    upgrade_db_parser.set_defaults(func=upgrade_db)

    recreate_discussion_parser = subparsers.add_parser('recreate-discussion',
                                                       help='Re-create discussion for given bug')
    recreate_discussion_parser.add_argument('bugid',
                                            metavar='<Bug ID>',
                                            help='Bug ID for which to re-create dicsussion')
    recreate_discussion_parser.set_defaults(func=recreate_discussion)

    update_discussion_parser = subparsers.add_parser('update-discussion',
                                                     help='Update discussion for a given ticket')
    update_discussion_parser.add_argument('ticket_id',
                                          metavar='<Ticket ID>',
                                          help='Ticket ID for which to update dicsussion')
    update_discussion_parser.set_defaults(func=update_discussion)

    close_inactive_discussions_parser = subparsers.add_parser('close-inactive-discussions',
        help="Close discussions of inactive releases (which haven't been processed yet)")
    close_inactive_discussions_parser.add_argument('--dryrun', action='store_true',
                                                   default=False,
                                                   help="Don't make any actual changes")
    close_inactive_discussions_parser.set_defaults(func=close_inactive_discussions)


    args = parser.parse_args()

    if not hasattr(args, 'func'):
        # if func attribute is missing, no command was specified, print help and exit
        parser.print_help()
        sys.exit(1)

    # run the requested command
    args.func(args)


if __name__ == '__main__':
    exit = main()
    if exit:
        sys.exit(exit)
