#
# api.py - RESTful API handlers
#
# Copyright 2013, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
#   Ilgiz Islamgulov <ilgizisl@gmail.com>
#

from flask import Blueprint, request

from blockerbugs import app, db
from blockerbugs.models.milestone import Milestone
from blockerbugs.models.update import Update
from blockerbugs.models.release import Release
from blockerbugs.models.bug import Bug
from blockerbugs.util import pagure_bot
from . import errors
from .utils import get_or_404, JsonResponse, SVGResponse, check_signature

api_v0 = Blueprint('api', __name__, url_prefix='/api/v0')

ACCEPTED_BUGTYPES = [
    'proposed_blocker', 'proposed_fe',
    'rejected_blocker', 'rejected_fe',
    'accepted_blocker', 'accepted_fe',
    'accepted_0day', 'accepted_prevrel',
    'prioritized'
]


@api_v0.errorhandler(errors.RestApiError)
def api_error_handler(e):
    return JsonResponse(e.to_dict(), e.http_status_code)


def get_update_info(update):
    update_simple_fields = ['title', 'url', 'karma', 'stable_karma', 'status',
                            'pending']
    update_data = dict(
        (attr, getattr(update, attr)) for attr in update_simple_fields)
    update_data['release'] = update.release.number
    update_data['milestones'] = [{'version': m.version,
                                  'release': m.release.number, }
                                 for m in update.milestones]
    update_data['bugs'] = [
        {'bugid': bug.bugid,
         'type': [tp for tp in ACCEPTED_BUGTYPES if getattr(bug, tp)]}
        for bug in update.bugs
    ]
    return update_data


def get_bug_info(bug):
    bug_simple_fields = ['bugid', 'url', 'summary', 'component', 'active']
    bug_info = dict((attr, getattr(bug, attr)) for attr in bug_simple_fields)
    bug_info['type'] = [tp for tp in ACCEPTED_BUGTYPES if getattr(bug, tp)]
    return bug_info

@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/updates')
def list_updates(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    m_alias = db.aliased(Milestone)
    updates = Update.query.join(m_alias, Update.milestones).filter(
        m_alias.id == milestone.id)
    if 'bugtype' in request.args:
        bugtype = request.args['bugtype']
        if bugtype in ACCEPTED_BUGTYPES:
            bug_attr = getattr(Bug, bugtype)
            updates = updates.filter(Update.bugs.any(bug_attr == True))
        else:
            raise errors.InvalidArgumentError(arg_name='bugtype')
    updates = updates.order_by(Update.date_submitted.desc()).all()
    updates_info = [get_update_info(up) for up in updates]
    return JsonResponse(updates_info)


@api_v0.route('/milestones/<int:rel_num>/<milestone_version>/bugs')
def list_bugs(rel_num, milestone_version):
    release = get_or_404(Release, number=rel_num)
    milestone = get_or_404(Milestone, release=release,
                           version=milestone_version)
    bugs = Bug.query.filter_by(milestone=milestone)
    if 'bugtype' in request.args:
        bugtype = request.args['bugtype']
        if bugtype in ACCEPTED_BUGTYPES:
            bug_attr = getattr(Bug, bugtype)
            bugs = bugs.filter(bug_attr == True)
        else:
            raise errors.InvalidArgumentError(arg_name='bugtype')
    bugs = bugs.order_by(Bug.component, Bug.bugid).all()
    bugs_info = [get_bug_info(bug) for bug in bugs]
    return JsonResponse(bugs_info)

@api_v0.route('/milestones/current')
def get_current_milestone():
    current_milestone = get_or_404(Milestone, current=True)
    return JsonResponse(current_milestone.simple())


@api_v0.route('/webhook', methods=['POST'])
def pagure_webhook():
    if not app.config['PAGURE_BOT_ENABLED']:
        msg = 'Pagure bot disabled, ignoring request'
        app.logger.info(msg)
        return JsonResponse({'msg': msg})

    if not check_signature(request.headers, request.data):
        msg = 'Wrong signature, ignoring.'
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})

    data = request.json
    if data['topic'] not in ['issue.comment.added', 'issue.comment.edited']:
        msg = 'Ignoring message with topic %s' % data['topic']
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})

    issue_id = data.get('msg', {}).get('issue', {}).get('id', None)
    status = data.get('msg', {}).get('issue', {}).get('status', '')
    if not issue_id or not status:
        msg = 'Unable to parse received message (isssue id, status)'
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})

    if status == 'Closed':
        msg = 'Ignoring a closed issue'
        app.logger.debug(msg)
        return JsonResponse({'msg': msg})

    pagure_bot.webhook_handler(issue_id)
    msg = 'Message successfully parsed'
    app.logger.debug(msg)
    return JsonResponse({'msg': msg})


def _svg_response_text(info_all):
    svg_template = '<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" height="{total_height}" width="300">'\
                   '{info_lines}'\
                   '</svg>'
    info_line_template = '<text font-family="sans-serif" x="0" y="{y_offset}" font-size="14px" fill="#212529">{info}</text>'

    # height of each line is 17px, text is rendered at y offset 13px (4px bellow as padding)
    # y_offset of nth line is 13 * (n + 1) + 4 * n = 13*n + 13 + 4*n = 17*n + 13
    info_lines = [info_line_template.format(y_offset=17 * index + 13,
                                            info=info) for index, info in enumerate(info_all)]
    return svg_template.format(total_height=17 * len(info_all), info_lines="".join(info_lines))


def _get_bugtypes(bug):
    bugtypes = []
    if bug.proposed_blocker:
        bugtypes.append("Proposed Blocker")
    if bug.proposed_fe:
        bugtypes.append("Proposed FE")
    if bug.rejected_blocker:
        bugtypes.append("Rejected Blocker")
    if bug.rejected_fe:
        bugtypes.append("Rejected FE")
    if bug.accepted_blocker:
        bugtypes.append("Accepted Blocker")
    if bug.accepted_0day:
        bugtypes.append("Accepted 0day")
    if bug.accepted_prevrel:
        bugtypes.append("Accepted Previous release")
    if bug.accepted_fe:
        bugtypes.append("Accepted FE")
    return bugtypes

_UNKNOWN_BUG_SVG_TEXT = _svg_response_text(["unknown bug"])
_BUG_CLOSED = "BUG CLOSED"

@api_v0.route('/bugimg/<int:bug_id>')
def bug_image(bug_id):
    bugs = Bug.query.filter_by(bugid=bug_id).all()
    if not bugs:
        return SVGResponse(_UNKNOWN_BUG_SVG_TEXT)
    else:
        info_all = []
        # same bug can be in DB multiple times (for different milestones),
        # only bugs in active milestones are updated, status of bug in
        # inactive milestone will remain CLOSED even if bug is reopened for
        # active milestone
        # show BUG CLOSED when all bug entries in DB have status == "CLOSED"
        if all([bug.status == "CLOSED" for bug in bugs]):
            info_all.append(_BUG_CLOSED)
        for bug in bugs:
            milestone_info = "%s: " % bug.milestone.name
            bugtypes = _get_bugtypes(bug)
            info_all.append(milestone_info + ", ".join(bugtypes))

        return SVGResponse(_svg_response_text(info_all))
