#!/usr/bin/python3

import datetime
import re
import collections

from blockerbugs import app
from blockerbugs.util import pagure_interface

TRACKER_KEYWORDS = [
    'betablocker',
    'finalblocker',
    'betafreezeexception',
    'finalfreezeexception',
    '0day',
    'previousrelease'
]

VOTES = ("+1", "0", "-1")
TRACKER_RE = r'((beta|final)(blocker|fe|freezeexception)|0day|previousrelease)'
TRACKER_MATCHER = re.compile(TRACKER_RE)
OUTCOME_RE = r'(accepted|rejected)' + TRACKER_RE
OUTCOME_MATCHER = re.compile(OUTCOME_RE)

SUMMARY_HEADER = ""
CLOSED_VOTES_HEADER = "The following votes have been closed:"

# Containers for parsed commands
VoteCommand = collections.namedtuple('VoteCommand', ['tracker', 'vote'])
VoteCommand.__doc__ += (
    ': A parsed vote command, e.g. tracker="betablocker", vote="+1"')

AgreedCommand = collections.namedtuple(
    'AgreedCommand', ['tracker', 'outcome', 'summary'], defaults=[None])
AgreedCommand.__doc__ += (
    ': A parsed AGREED command, e.g. tracker="betablocker", '
    'outcome="accepted", summary="Breaks boot."')

RevoteCommand = collections.namedtuple('RevoteCommand', ['tracker'])
RevoteCommand.__doc__ += ': A parsed REVOTE command, e.g. tracker="betablocker"'


def agreed_revote_parser(line):
    """Parse an AGREED or a REVOTE line. If there's anything violating the
    rules for the line (e.g. it doesn't start with the expected command, or
    there's some extra text that shouldn't be there), ignore the line contents
    and return []. All instances of "fe" get expanded to "freezeexception".

    :param str line: a single line
    :return: A list of :class:`AgreedCommand` or :class:`RevoteCommand` instances.
    E.g. for the line 'agreed acceptedbetablocker rejectedfinalfe' this returns:
    [AgreedCommand(tracker='betablocker', outcome='accepted'),
     AgreedCommand(tracker='finalfreezeexception', outcome='rejected')]
    For the line 'revote finalblocker 0day' this returns:
    [RevoteCommand(tracker='finalblocker'),
     RevoteCommand(tracker='0day')]
    """
    # parsing word by word avoids complex and fragile regex syntax
    words = line.split()
    if len(words) <= 0:
        return []

    out = []
    command = words[0]
    for word in words[1:]:
        if command == 'agreed':
            match = OUTCOME_MATCHER.fullmatch(word)
            if (match):
                ac = AgreedCommand(
                    tracker=expand_fe(match.group(2)),
                    outcome=match.group(1),
                )
                out.append(ac)
            else:
                # this must be the summary now
                # FIXME: implement summary parsing
                break
        elif command == 'revote':
            match = TRACKER_MATCHER.fullmatch(word)
            if (match):
                rc = RevoteCommand(
                    tracker=expand_fe(word),
                )
                out.append(rc)
            else:
                # this is not a valid REVOTE line
                return []
        else:
            # not an AGREED or REVOTE line
            return []
    return out


def vote_parser(line):
    """Look for votes in a single line. We accept pairs of a vote and a match
    of TRACKER_RE, in either order, and nothing else; any other text
    invalidates the line. All instances of "fe" get expanded to "freezeexception".

    :param str line: a line which might contain a vote
    :return: A list of :class:`VoteCommand` instances.
    E.g. for the line 'betafe +1 finalblocker -1' this returns:
    [VoteCommand(tracker='betafreezeexception', vote='+1'),
     VoteCommand(tracker='finalblocker', vote='-1')]
    """
    words = line.split()
    # if we have an odd number of words we don't have a clean vote line
    if (len(words) % 2) != 0:
        return []

    out = []
    # split list into pairs:
    # https://stackoverflow.com/questions/312443
    for pair in (words[i:i + 2] for i in range(0, len(words), 2)):
        if TRACKER_MATCHER.fullmatch(pair[0]) and pair[1] in VOTES:
            vc = VoteCommand(
                tracker=expand_fe(pair[0]),
                vote=pair[1]
            )
            out.append(vc)
        elif pair[0] in VOTES and TRACKER_MATCHER.fullmatch(pair[1]):
            vc = VoteCommand(
                tracker=expand_fe(pair[1]),
                vote=pair[0]
            )
            out.append(vc)
        else:
            # we found something other than a vote pair, so per
            # Kamil's Strict Vote Parsing Regime, we reject the line
            return []
    return out


def expand_fe(tracker):
    '''Expand "fe" into "freezeexception" in tracker names.'''
    if tracker == 'betafe':
        return 'betafreezeexception'
    elif tracker == 'finalfe':
        return 'finalfreezeexception'
    else:
        return tracker


def bot_loop_detected(comments):
    '''Detect a loop of bot comments in the list of comments. The loop is
    defined as a continuous stream of more than $threshold comments, all from
    the bot, including the last comment in the list. The threshold is defined
    in config.

    :param comments: a list of Comment instances
    :return: bool whether there is a loop
    '''
    is_bot_comment = [comment.user == app.config['PAGURE_BOT_USERNAME'] for comment in comments]
    threshold = app.config['PAGURE_BOT_LOOP_THRESHOLD']
    if len(is_bot_comment) < threshold:
        return False

    return all(is_bot_comment[-threshold:])


class Comment():
    '''This represents a single comment from the voting ticket.
    '''
    def __init__(self, raw_comment):
        self.text = raw_comment['comment'].replace('\r', '')
        self.user = raw_comment['user']['name']
        #: Pagure's comment id (int)
        self.id = raw_comment['id']

    def __repr__(self):
        return "%s (%s): %s" % (self.user, self.id, self.text)

    def user_is_admin(self):
        # FIXME use config for group name
        # FIXME it would be good to save/cache the data
        return self.user in pagure_interface.get_group('fedora-qa')['members']

    def commands(self):
        '''Search for commands in this comment, purge anything unrelated and
        return it parsed.

        :return: a list of :class:`VoteCommand`, :class:`AgreedCommand`, and
            :class:`RevoteCommand` instances, e.g.:
            [VoteCommand(tracker='betablocker', vote='+1'),
             AgreedCommand(tracker='betablocker', outcome='accepted'),
             RevoteCommand(tracker='betablocker'),
             VoteCommand(tracker='betablocker', vote='-1')]
        '''
        out = []
        for line in self.text.lower().split('\n'):
            line = line.strip()
            if line.startswith('agreed') or line.startswith('revote'):
                out.extend(agreed_revote_parser(line))
            else:
                out.extend(vote_parser(line))
        return out

    def is_summary_post_of(self, tracker_keyword):
        '''Return True/False whether this particular post is a summary of
        voting results for a particular tracker.
        '''
        is_bot = (self.user == app.config['PAGURE_BOT_USERNAME'])
        is_summary = (CLOSED_VOTES_HEADER.lower() in self.text.lower())
        is_tracker_relevant = (tracker_keyword in self.text.lower())

        return is_bot and is_summary and is_tracker_relevant


class BugVoteTracker():
    '''Count votes specific for a single tracker. Just feed all comments to
    this class and it'll reflect the tracker's final state in its variables.
    '''

    def __init__(self, tracker):
        self.open = True
        #: 'accepted' or 'rejected'
        self.outcome = None
        #: one of TRACKER_KEYWORDS
        self.tracker = tracker
        self.need_summary_post = False

        #: user votes in this format:
        #: {
        #:     'user': {
        #:          'vote': '+1'
        #:          'comment_id': 21383
        #:     }
        #: }
        self.votes = {}

    def parse_comment(self, comment):
        '''Take a comment and parse all commands from it that are relevant to
        this particular tracker. Update the instance variables.'''

        if comment.is_summary_post_of(self.tracker):
            self.need_summary_post = False
            return

        # now that we detected summary posts, we can ignore all posts from the bot
        if comment.user == app.config['PAGURE_BOT_USERNAME']:
            return

        for command in comment.commands():
            if command.tracker != self.tracker:
                # not intended for this tracker, ignore
                continue

            agreed = isinstance(command, AgreedCommand)
            revote = isinstance(command, RevoteCommand)
            vote = isinstance(command, VoteCommand)

            if (agreed or revote) and not comment.user_is_admin():
                app.logger.debug(
                    f'A non-admin user {comment.user} tries to perform '
                    f'administrative commands in comment {comment.id}, '
                    'ignoring.')
                continue

            if agreed:
                self.open = False
                self.outcome = command.outcome
                self.need_summary_post = True
            elif revote:
                self.open = True
                self.outcome = None
                self.need_summary_post = False
                self.votes = {}
            elif vote:
                if not self.open:
                    continue
                self.votes[comment.user] = {
                    'vote': command.vote,
                    'comment_id': comment.id
                }
            else:
                msg = f'Unknown command instance: {command}'
                app.logger.error(msg)
                assert False, msg

    def enumerate_votes(self):
        '''Return self.votes in a format:
        { '-1': [('user1', comment_id), ('user2', comment_id)],
           '0': [...],
          '+1': [...],
        }
        '''
        out = {vote: [] for vote in VOTES}
        for user, vote in self.votes.items():
            out[vote['vote']].append((user, vote['comment_id']))

        return out


def link(user, comment_id):
    '''Create a Markdown snippet containing a link to the user's comment.
    '''
    return "[%s](#comment-%s)" % (user, comment_id)


def tracker_summary(tracker_name, tracker):
    '''Create a summary of votes for a given tracker in a Markdown format.

    :param str tracker_name: e.g. 'BetaBlocker'
    :param tracker: an instance of BugVoteTracker
    '''
    nicer = {
        'betablocker': 'BetaBlocker',
        'finalblocker': 'FinalBlocker',
        'betafreezeexception': 'BetaFreezeException',
        'finalfreezeexception': 'FinalFreezeException',
        '0day': '0Day',
        'previousrelease': 'PreviousRelease',
        'accepted': 'Accepted',
        'rejected': 'Rejected'
    }

    out = "* "

    outcome = ""
    if not tracker.open:
        outcome = "**%s** " % nicer[tracker.outcome]

    out += "%s%s " % (outcome, nicer[tracker_name])

    votes = tracker.enumerate_votes()
    pros = len(votes['+1'])
    neutrals = len(votes['0'])
    cons = len(votes['-1'])

    pros_text = "+%s" % pros
    if pros > 0:
        pros_text = "**+%s**" % pros

    cons_text = "-%s" % cons
    if cons > 0:
        cons_text = "**-%s**" % cons

    out += "(%s, %s, %s)\n" % (pros_text, neutrals, cons_text)

    return out


def summary(trackers, non_voting_users=None, last_comment_id=None, header=CLOSED_VOTES_HEADER):
    '''Create a summary text in a Markdown format which can be used either in
    a ticket description, or as a comment when a vote is closed.

    :param trackers: a dict as returned from :func:`parse_comments_to_trackers`
    :param non_voting_users: a list of users who haven't voted yet
    :param int last_comment_id: the last comment id that was counted
    :param str head: an initial message text
    '''
    md_text = "%s\n\n" % header

    for tracker_name, tracker in trackers.items():
        if not tracker.votes:
            if not tracker.open:
                md_text += tracker_summary(tracker_name, tracker)
            continue

        votes = tracker.enumerate_votes()

        md_text += tracker_summary(tracker_name, tracker)

        for vote in VOTES:
            votes[vote] = [link(user, comment_id) for (user, comment_id) in votes[vote]]
            if len(votes[vote]) == 0:
                continue
            md_text += "    * %s by %s\n" % (vote, ', '.join(votes[vote]))

    md_text += '\n'

    if non_voting_users:
        md_text += "Commented but haven't voted yet: %s\n\n" % ', '.join(
            [link(u, c) for (u, c) in non_voting_users.items()])

    if last_comment_id:
        time = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M")

        md_text += ("*The votes have been last counted at %s UTC and the last "
                    "processed comment was [#comment-%s](#comment-%s)*" % (
                        time, last_comment_id, last_comment_id))

    return md_text


def parse_comments_to_trackers(comments):
    '''Create BugVoteTracker for each tracker keyword supported, and then feed
    all comments to each tracker to fully update them.

    :param comments: a list of Comment instances
    :return: A dict in format:
            {'betablocker': BugVoteTracker,
             'betafreezeexception': BugVoteTracker,
             ...
            }
    '''
    trackers = {tracker: BugVoteTracker(tracker) for tracker in TRACKER_KEYWORDS}

    for comment in comments:
        for tracker in trackers.values():
            tracker.parse_comment(comment)

    return trackers


def voting_info(comments, trackers):
    '''Compute users who haven't voted yet, and the last processed comment

    :param comments: a list of Comment instances
    :param trackers: a dict as returned from :func:`parse_comments_to_trackers`
    :return: a tuple of (dict of non_voting_users, last_comment_id), e.g.:
    ({'user1': 1234,
      'user2': 1235
     },
     1238)
    '''
    last_comment_id = 0
    non_voting_users = {}
    voting_users = set()

    for comment in comments:
        last_comment_id = comment.id
        if comment.user != app.config['PAGURE_BOT_USERNAME']:
            non_voting_users[comment.user] = comment.id

    for tracker in trackers.values():
        voting_users |= tracker.votes.keys()

    for user in voting_users:
        non_voting_users.pop(user)

    return non_voting_users, last_comment_id


def webhook_handler(issue_id):
    '''React to a Pagure webhook notification that a discussion ticket changed.
    '''
    comments = [Comment(c) for c in pagure_interface.get_issue_comments(issue_id)]

    trackers = parse_comments_to_trackers(comments)

    non_voting_users, last_comment_id = voting_info(comments, trackers)

    trackers_without_summary = {n: t for (n, t) in trackers.items() if t.need_summary_post}
    if trackers_without_summary:
        app.logger.debug('Agreed/Rejected keyword found, posting summary comment')
        if bot_loop_detected(comments):
            app.logger.error('Bot is probably stuck in loop, not posting summary comment!')
        else:
            comment = summary(trackers_without_summary)
            pagure_interface.post_comment(issue_id, comment)

    app.logger.debug('Updating issue summary')
    vote_summary = summary(trackers, non_voting_users, last_comment_id, SUMMARY_HEADER)
    pagure_interface.update_issue(issue_id, vote_summary)
